import copy

from lxml import etree

from six.moves import zip_longest


class Base(object):
    _require_keyword_arg = False

    @property
    def is_optional(self):
        return self.min_occurs == 0

    def signature(self, name=None):
        return ''


class Any(Base):
    _require_keyword_arg = False

    def __init__(self, max_occurs=1, min_occurs=1):
        self.name = 'any'
        self.max_occurs = max_occurs
        self.min_occurs = min_occurs

        # cyclic import
        from zeep.xsd.builtins import AnyType
        self.type = AnyType()

    def __repr__(self):
        return '<%s(name=%r)>' % (self.__class__.__name__, self.name)

    def render(self, parent, value):
        assert parent is not None
        if value is None:
            # not possible
            return

        if isinstance(value.value, list):
            for val in value.value:
                value.xsd_type.render(parent, val)
        else:
            value.xsd_type.render(parent, value.value)

    def parse(self, xmlelement, schema):
        xsd_type = xmlelement.get('{http://www.w3.org/2001/XMLSchema-instance}type')
        if xsd_type is not None:
            element_type = schema.get_type(xsd_type)
            return element_type.parse(xmlelement, schema)

        try:
            element_type = schema.get_element(xmlelement.tag)
            return element_type.parse(xmlelement, schema)
        except KeyError:
            return xmlelement

    def __call__(self, any_object):
        return any_object

    def signature(self, name=None):
        return '%s%s: %s' % (
            name, '=None' if self.is_optional else '',
            '[]' if self.max_occurs != 1 else ''
        )


class Element(Base):
    def __init__(self, name, type_=None, min_occurs=1, max_occurs=1,
                 nillable=False):
        if name and not isinstance(name, etree.QName):
            name = etree.QName(name)

        self.name = name.localname if name else None
        self.qname = name
        self.type = type_
        self.min_occurs = min_occurs
        self.max_occurs = max_occurs
        self.nillable = nillable
        # assert type_

    def __str__(self):
        if self.type:
            return '%s(%s)' % (self.name, self.type.signature())
        return '%s()' % self.name

    def __call__(self, *args, **kwargs):
        instance = self.type(*args, **kwargs)
        if hasattr(instance, '_xsd_type'):
            instance._xsd_elm = self
        return instance

    def __repr__(self):
        return '<%s(name=%r, type=%r)>' % (
            self.__class__.__name__, self.name, self.type)

    def __eq__(self, other):
        return (
            other is not None and
            self.__class__ == other.__class__ and
            self.__dict__ == other.__dict__)

    def signature(self, name=None):
        assert self.type, '%r has no name' % self
        return '%s%s: %s%s' % (
            name, '=None' if self.is_optional else '',
            self.type.name, '[]' if self.max_occurs != 1 else ''
        )

    def clone(self, name):
        if not isinstance(name, etree.QName):
            name = etree.QName(name)

        new = copy.copy(self)
        new.name = name.localname
        new.qname = name
        return new

    def serialize(self, value):
        return self.type.serialize(value)

    def resolve_type(self, schema):
        self.type = self.type.resolve(schema)

    def render(self, parent, value):
        assert parent is not None

        if value is None:
            if not self.is_optional:
                node = etree.SubElement(parent, self.qname)
            return

        if self.name is None:
            return self.type.render(parent, value)

        node = etree.SubElement(parent, self.qname)

        xsd_type = getattr(value, '_xsd_type', self.type)
        if xsd_type != self.type:
            return value._xsd_type.render(node, value, xsd_type)
        return self.type.render(node, value)

    def parse(self, value, schema):
        return self.type.parse_xmlelement(value, schema)


class Attribute(Element):
    def __init__(self, name, type_=None, required=False, default=None):
        super(Attribute, self).__init__(name=name, type_=type_)
        self.required = required
        self.default = default or ''

    def render(self, parent, value):
        if value is None:
            if self.default:
                value = self.default
            elif not self.required:
                return
            else:
                value = ""  # XXX Throw exception?

        value = self.type.xmlvalue(value)
        parent.set(self.qname, value)

    def parse(self, value, schema=None):
        return self.type.pythonvalue(value)


class ListElement(Element):

    def __call__(self, *args, **kwargs):
        return [self.type(*args, **kwargs)]

    def serialize(self, value):
        if value:
            return [self.type.serialize(val) for val in value]
        return []

    def render(self, parent, value):
        for val in value:
            super(ListElement, self).render(parent, val)


class GroupElement(Element):
    def __init__(self, *args, **kwargs):
        self.children = kwargs.pop('children', [])
        assert self.children
        assert isinstance(self.children, list)
        super(GroupElement, self).__init__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        for item in self.children:
            yield item

    def properties(self):
        return self.children

    def signature(self, name):
        return '%s%s: %s' % (
            name, '=None' if self.is_optional else '',
            '[]' if self.max_occurs != 1 else ''
        )


class Choice(Base):
    _require_keyword_arg = False

    def __init__(self, elements, max_occurs=1, min_occurs=1):
        self.name = 'choice'
        self.type = None
        self.elements = elements
        self.max_occurs = max_occurs
        self.min_occurs = min_occurs

    @property
    def is_optional(self):
        return True

    def key(self):
        # XXX Any elemetns?
        return ':'.join(elm.name for elm in self.elements)

    def render(self, parent, name, value):
        choice_metadata = getattr(value, name)

        if self.max_occurs == 1:
            choice_metadata = [choice_metadata]

        for item in choice_metadata:
            choice_element = self.elements[item.index]

            if isinstance(choice_element, Element):
                choice_value = item.values.get(choice_element.name)
                if choice_value is None and self.is_optional:
                    return

                if isinstance(choice_value, list):
                    for item in choice_value:
                        choice_element.render(parent, item)
                else:
                    choice_element.render(parent, choice_value)
            else:
                for element in choice_element:
                    value = item.values[element.name]
                    if isinstance(value, list):
                        for item in value:
                            element.render(parent, item)
                    else:
                        element.render(parent, value)

    def signature(self, name):
        part = ' | '.join([
            '{%s}' % element.signature(element.name)
            for element in self.elements
        ])

        if self.max_occurs != 1:
            return '%s: [%s]' % (name, part)
        return '%s: %s' % (name, part)

    def parse(self, elements, schema):
        result = []
        i = 0
        for choice_element in elements:
            for iter_field in self.elements:
                if isinstance(iter_field, Sequence):
                    subresult = iter_field.parse(elements[i:], schema)
                    if subresult:
                        i += len(subresult)
                        result.append(subresult)
                        break
                else:
                    if iter_field.qname == choice_element.tag:
                        choice_result = iter_field.parse(choice_element, schema)
                        result.append({
                            iter_field.name: choice_result
                        })
                        i += 1
                        break

            if len(result) >= self.max_occurs:
                break

        return result


class Sequence(list):
    name = 'sequence'

    def signature(self, name):
        return ', '.join([
            element.signature(element.name) for element in self
        ])

    def parse(self, elements, schema):

        result = {}
        for field, element in zip_longest(self, elements):
            if field is None:
                break

            if element is None and field.is_optional:
                result[field.name] = None
            elif field.qname == element.tag:
                result[field.name] = field.parse(element, schema)
            elif field.is_optional:
                result[field.name] = None
            else:
                return None
        return result


class RefElement(object):

    def __init__(self, tag, ref, schema):
        self._ref = ref
        self._schema = schema

    @property
    def _elm(self):
        return self._schema.get_element(self._ref)

    def __iter__(self, *args, **kwargs):
        elm = self._elm

        if isinstance(elm, (GroupElement, ListElement)):
            for item in elm.properties():
                yield item
        else:
            yield elm

    def __call__(self, *args, **kwargs):
        return self._elm(*args, **kwargs)

    def __getattr__(self, name):
        if not name.startswith('_'):
            return getattr(self._elm, name)

        return getattr(self, name)


class RefAttribute(RefElement):

    @property
    def _elm(self):
        return self._schema.get_attribute(self._ref)



import itertools

from collections import OrderedDict

import six

from zeep.xsd.elements import (
    Any, Attribute, Choice, Element, GroupElement, ListElement, RefElement,
    Sequence)
from zeep.xsd.valueobjects import CompoundValue


class Type(object):

    def accept(self, value):
        raise NotImplementedError

    def parse_xmlelement(self, xmlelement, schema=None):
        raise NotImplementedError

    def parsexml(self, xml, schema=None):
        raise NotImplementedError

    def render(self, parent, value):
        raise NotImplementedError

    def resolve(self, schema):
        raise NotImplementedError

    @classmethod
    def signature(cls):
        return ''


class UnresolvedType(Type):
    def __init__(self, qname):
        self.qname = qname

    def resolve(self, schema):
        retval = schema.get_type(self.qname)
        return retval.resolve(schema)


class UnresolvedCustomType(Type):

    def __init__(self, name, base_qname):
        assert name is not None
        self.name = name
        self.base_qname = base_qname

    def resolve(self, schema):
        base = schema.get_type(self.base_qname)
        base = base.resolve(schema)

        cls_attributes = {
            '__module__': 'zeep.xsd.dynamic_types',
        }
        xsd_type = type(self.name, (base.__class__,), cls_attributes)
        return xsd_type()


@six.python_2_unicode_compatible
class SimpleType(Type):
    name = None

    def __eq__(self, other):
        return (
            other is not None and
            self.__class__ == other.__class__ and
            self.__dict__ == other.__dict__)

    def render(self, parent, value):
        parent.text = self.xmlvalue(value)

    def parse_xmlelement(self, xmlelement, schema=None):
        if xmlelement.text is None:
            return
        return self.pythonvalue(xmlelement.text)

    def xmlvalue(self, value):
        raise NotImplementedError(
            '%s.xmlvalue() not implemented' % self.__class__.__name__)

    def pythonvalue(self, xmlvalue):
        raise NotImplementedError(
            '%s.pytonvalue() not implemented' % self.__class__.__name__)

    def resolve(self, schema):
        return self

    def serialize(self, value):
        return value

    def __call__(self, *args, **kwargs):
        """Return the xmlvalue for the given value.

        The args, kwargs handling is done here manually so that we can return
        readable error messages instead of only '__call__ takes x arguments'

        """
        num_args = len(args) + len(kwargs)
        if num_args != 1:
            raise TypeError((
                '%s() takes exactly 1 argument (%d given). ' +
                'Simple types expect only a single value argument'
            ) % (self.__class__.__name__, num_args))

        if kwargs and 'value' not in kwargs:
            raise TypeError((
                '%s() got an unexpected keyword argument %r. ' +
                'Simple types expect only a single value argument'
            ) % (self.__class__.__name__, next(six.iterkeys(kwargs))))

        value = args[0] if args else kwargs['value']
        return self.xmlvalue(value)

    def __str__(self):
        return self.name

    @classmethod
    def signature(cls):
        return 'value'


class ComplexType(Type):
    name = None

    def __init__(self, children=None):
        self._children = children or []

    def __call__(self, *args, **kwargs):
        if not hasattr(self, '_value_class'):
            self._value_class = type(
                self.__class__.__name__, (CompoundValue,),
                {'_xsd_type': self, '__module__': 'zeep.objects'})

        return self._value_class(*args, **kwargs)

    def properties(self):
        return list(self._children)

    def fields(self):
        """Return a list of tuples containing the name and element of the
        fields.

        """
        result = []
        num_any = 1
        num_choice = 1
        for prop in self._children:
            if isinstance(prop, Any):
                result.append(('_any_%d' % num_any, prop))
                num_any += 1
            elif isinstance(prop, Choice):
                result.append(('_choice_%d' % num_choice, prop))
                num_choice += 1
            elif prop.name is None:
                result.append(('_value', prop))
            else:
                result.append((prop.name, prop))
        return result

    def serialize(self, value):
        return OrderedDict([
            (field.name, field.serialize(getattr(value, field.name, None)))
            for field in self.properties()
        ])

    def render(self, parent, value, xsd_type=None):
        for name, element in self.fields():
            if isinstance(element, Choice):
                element.render(parent, name, value)
            else:
                sub_value = getattr(value, name, None)
                element.render(parent, sub_value)

        if xsd_type:
            parent.set(
                '{http://www.w3.org/2001/XMLSchema-instance}type',
                xsd_type._xsd_name)

    def resolve(self, schema):
        children = []
        for elm in self._children:
            if isinstance(elm, RefElement):
                elm = elm._elm

            if isinstance(elm, UnresolvedType):
                xsd_type = elm.resolve(schema)
                if isinstance(xsd_type, SimpleType):
                    children.append(Element(None, xsd_type))
                else:
                    children.extend(list(xsd_type._children))

            elif isinstance(elm, GroupElement):
                children.extend(list(elm))
            else:
                children.append(elm)
        self._children = children
        return self

    def signature(self):
        parts = []

        for name, element in self.fields():
            part = element.signature(name)
            parts.append(part)
        return ', '.join(parts)

    def parse_xmlelement(self, xmlelement, schema):
        instance = self()
        fields = self.fields()
        if not fields:
            return instance

        elements = xmlelement.getchildren()
        attributes = xmlelement.attrib
        if not elements and not attributes:
            return

        fields_map = {k: v for k, v in fields if isinstance(v, Attribute)}
        for key, value in attributes.items():
            field = fields_map.get(key)
            if not field:
                continue
            value = field.parse(value, schema)
            setattr(instance, key, value)

        fields = iter((k, v) for k, v in fields if not isinstance(v, Attribute))
        field_name, field = next(fields, (None, None))

        # If the type has no child elements (only attributes) then return
        # early
        if not field:
            return instance

        i = 0
        num_elements = len(elements)
        while i < num_elements:
            element = elements[i]
            result = None

            # Find matching element
            while field:
                if isinstance(field, (Choice, Any)):
                    break

                if field.qname == element.tag:
                    break

                field_name, field = next(fields, (None, None))

            if isinstance(field, Choice):
                result = field.parse(elements[i:], schema)
                i += sum(len(choice) for choice in result)

                # If the field has maxOccurs = 1 and is not a sequence then
                # make the interface easier to use by setting the property
                # directly.
                if field.max_occurs == 1:
                    if len(result[0]) == 1:
                        setattr(
                            instance,
                            next(six.iterkeys(result[0])),
                            next(six.itervalues(result[0])))

                setattr(instance, field_name, result)

            elif isinstance(field, Any):
                result = field.parse(element, schema)
                setattr(instance, field_name, result)
                i += 1

            else:
                if not field:
                    raise ValueError("Unexpected element: %r" % element)
                    break

                # Element can be optional, so if this doesn't match then assume
                # it was.
                if field.qname != element.tag:
                    continue

                current_field = field

                result = current_field.parse(element, schema)
                i += 1

                if isinstance(field, ListElement):
                    assert getattr(instance, field.name) is not None
                    getattr(instance, field_name).append(result)
                else:
                    setattr(instance, field_name, result)

            # Check if the next element also applies to the current field
            try:
                if field.max_occurs == 1 or element.tag != elements[i].tag:
                    field_name, field = next(fields, (None, None))
            except IndexError:
                break

        return instance

    @property
    def name(self):
        return self.__class__.__name__

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__, self.signature())


class ListType(Type):
    def __init__(self, item_type):
        self.item_type = item_type

    def resolve(self, schema):
        self.item_type = self.item_type.resolve(schema)
        return self

    def xmlvalue(self, value):
        item_type = self.item_type
        return ' '.join(item_type.xmlvalue(v) for v in value)

    def render(self, parent, value):
        parent.text = self.xmlvalue(value)


class UnionType(object):
    def __init__(self, item_types):
        self.item_types = item_types



import copy
import pprint
from collections import OrderedDict

import six

from zeep.xsd.elements import Any, Choice, ListElement, Sequence

__all__ = ['AnyObject', 'CompoundValue']


class AnyObject(object):
    def __init__(self, xsd_type, value):
        self.xsd_type = xsd_type
        self.value = value

    def __repr__(self):
        return '<%s(type=%r, value=%r)>' % (
            self.__class__.__name__, self.xsd_type, self.value)


class CompoundValue(object):

    def __init__(self, *args, **kwargs):
        fields = self._xsd_type.fields()

        # Set default values
        for key, value in fields:
            if isinstance(value, ListElement):
                value = []
            else:
                value = None
            setattr(self, key, value)

        items = _process_signature(fields, args, kwargs)
        fields = OrderedDict([(name, elm) for name, elm in fields])
        for key, value in items.items():

            if key in fields:
                field = fields[key]
                value = _convert_value(field, value)

            setattr(self, key, value)

    def __repr__(self):
        return pprint.pformat(self.__dict__, indent=4)


class ChoiceItem(object):
    def __init__(self, index, values):
        self.index = index
        self.values = values

    def __repr__(self):
        return '<%s(index=%r, values=%r)>' % (
            self.__class__.__name__, self.index, self.values)

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__ and
            self.__dict__ == other.__dict__)

    def __getitem__(self, key):
        return self.values[key]

    def __setitem__(self, key, value):
        self.values[key] = value


def _process_signature(fields, args, kwargs):
    """Return a dict with the args/kwargs mapped to the field name.

    Special handling is done for Choice elements since we need to record which
    element the user intends to use.

    :param fields: List of tuples (name, element)
    :type fields: list
    :param args: arg tuples
    :type args: tuple
    :param kwargs: kwargs
    :type kwargs: dict


    """
    result = {}

    all_fields = fields
    fields = list(fields)
    args = list(args)
    kwargs = copy.copy(kwargs)

    # Create a list of fields until field in any or choice

    num_pos_args = 0
    for name, element in fields:
        if element._require_keyword_arg:
            break
        num_pos_args += 1

    if len(args) > num_pos_args:
        raise TypeError(
            "__init__() takes at most %s positional arguments (%s given)" % (
                num_pos_args, len(args)))

    # Process the positional arguments
    for name, element in list(fields):
        if not args:
            break
        arg = args.pop(0)
        if element._require_keyword_arg:
            raise TypeError(
                "Any and Choice elements should be passed using a keyword argument")
        result[name] = arg
        fields.pop(0)
    if args:
        assert False

    # Process the keyword arguments
    for i, (name, element) in enumerate(fields):
        if isinstance(element, Choice):
            result[name] = _process_signature_choice(name, element, kwargs)

        elif name in kwargs:
            value = kwargs.pop(name)

            if isinstance(element, Any) and not isinstance(value, AnyObject):
                raise TypeError(
                    "%s: expected AnyObject, %s found" % (
                        name, type(value).__name__))

            result[name] = value

    if kwargs:
        raise TypeError(
            (
                "__init__() got an unexpected keyword argument %r, " +
                "Valid keyword arguments are: %s"
            ) % (next(six.iterkeys(kwargs)), ', '.join(x[0] for x in all_fields)))

    return result


def _process_signature_choice(name, element, kwargs):
    """Processes the kwargs for given choice element.

    Returns a list with multiple `utils.ChoiceItem` objects if maxOccurs > 1
    or simply one object if maxOccurs = 0.

    This handles two distinct initialization methods:

    1. Passing the choice elements directly to the kwargs (unnested)
    2. Passing the choice elements into the `name` kwarg (_choice_1) (nested).
       This case is required when multiple choice elements are given.

    :param name: Name of the choice element (_choice_1)
    :type name: str
    :param element: Choice element object
    :type element: zeep.xsd.Choice
    :param kwargs: dict (or list of dicts) of kwargs for initialization
    :type kwargs: list / dict

    """
    result = []
    if name in kwargs:
        values = kwargs.pop(name)
        if isinstance(values, dict):
            values = [values]

        for value in values:
            for choice_index, choice in enumerate(element.elements):

                if isinstance(choice, Sequence):
                    match = all(
                        child.name in value for child in choice
                        if not child.is_optional
                    )

                    if match:
                        result.append(ChoiceItem(choice_index, value))
                        break
                else:
                    if choice.name in value:
                        result.append(ChoiceItem(choice_index, value))
                        break
            else:
                raise TypeError(
                    "No complete xsd:Sequence found for the xsd:Choice %r.\n"
                    "The signature is: %s" % (name, element.signature(name)))

    else:

        # When choice elements are specified directly in the kwargs
        for choice_index, choice in enumerate(element.elements):

            if isinstance(choice, Sequence):
                match = all(
                    child.name in kwargs for child in choice
                    if not child.is_optional)

                if match:
                    item_values = {}
                    for child in choice:
                        if child.is_optional and child.name not in kwargs:
                            item_values[child.name] = None
                        else:
                            item_values[child.name] = kwargs.pop(child.name)
                    result.append(ChoiceItem(choice_index, item_values))
                    break
                elif any(
                    child.name in kwargs for child in choice
                    if not child.is_optional
                 ):
                    raise TypeError(
                        "No complete xsd:Choice %r.\n"
                        "The signature is: %s" % (name, element.signature(name)))

            else:
                if choice.name in kwargs:
                    result.append(
                        ChoiceItem(
                            choice_index,
                            {choice.name: kwargs.pop(choice.name)})
                    )

            if len(result) >= element.max_occurs:
                break

    if element.max_occurs == 1:
        if len(result) == 1:
            return result[0]
        elif len(result) > 1:
            raise TypeError("Number of items is larger then max_occurs")
        else:
            return ChoiceItem(0, {})
    return result


def _convert_value(field, value):

    if isinstance(field, Choice):
        return value

    if isinstance(value, dict):
        value = field(**value)

    elif isinstance(value, list):
        value = [_convert_value(field.type, v) for v in value]

    return value
